from typing import Any, Dict, List, Optional, Union

from .plugin import Plugin
from .step import Step
from .retry import AutomaticRetry, ManualRetry

from ..util import generate_key


class Command(Step):
    """
    A Step which runs an executable in the repository and marks the step as
    pass/fail based on exit code.

    See: https://buildkite.com/docs/pipelines/command-step
    """

    def __init__(
        self,
        label: str,
        command: str,
        *,
        ## sort alphabetically:
        agents: Optional[dict] = None,
        artifact_paths: Union[str, List[str]] = "",
        automatic_retries: Union[None, int, List[AutomaticRetry]] = None,
        concurrency: int = 0,
        concurrency_group: str = "",
        env: Optional[dict] = None,
        key: str = "",
        manual_retry: Optional[ManualRetry] = None,
        parallelism: int = 0,
        plugins: Optional[List[Plugin]] = None,
        priority: Optional[int] = None,
        skip_reason: str = "",
        soft_fail: Union[bool, list] = False,
        tags: List[str] = None,
        timeout_in_minutes: int = 0,
        **kwargs
    ):
        super().__init__(tags=tags, **kwargs)
        self.key = key or generate_key(label)
        self.label = label
        self.command = command
        self.env = env or {}
        self.agents = agents or {}
        self.automatic_retries: List[AutomaticRetry] = []
        self.soft_fail = soft_fail
        if automatic_retries:
            if isinstance(automatic_retries, int):
                self.automatic_retries = [AutomaticRetry(automatic_retries)]
            else:
                self.automatic_retries = list(automatic_retries)
        self.manual_retry = manual_retry
        self.concurrency = concurrency
        self.concurrency_group = concurrency_group
        self.plugins = plugins
        self.priority = priority
        self.artifact_paths: List[str] = []
        if artifact_paths:
            if isinstance(artifact_paths, str):
                self.artifact_paths = artifact_paths.split(";")
            else:
                self.artifact_paths = artifact_paths
        self.skip_reason = skip_reason
        self.parallelism = parallelism
        self.timeout_in_minutes = timeout_in_minutes

    def asdict(self) -> dict:
        d = super().asdict()
        d["command"] = self.command
        d["label"] = self.label
        d["key"] = self.key

        # Setup inheritable properties using class-based defaults using MRO
        # to aggregate hash and list types, or determine the first valid value
        # for scalar types like timeout_in_minutes
        env: dict[str, str] = {}
        agents: dict[str, str] = {}
        artifact_paths = set(self.artifact_paths or [])
        plugins = set(self.plugins or [])
        timeout_in_minutes = self.timeout_in_minutes
        for cls in self.classes():
            env.update(getattr(cls, "env", {}))
            agents.update(getattr(cls, "agents", {}))
            artifact_paths |= set(getattr(cls, "artifact_paths", []))
            plugins |= set(getattr(cls, "plugins", []))
            if not timeout_in_minutes:
                timeout_in_minutes = getattr(cls, "timeout_in_minutes", 0)
        env.update(self.env)
        agents.update(self.agents)

        if env:
            d["env"] = env
        if agents:
            d["agents"] = agents
        if artifact_paths:
            d["artifact_paths"] = list(sorted(list(artifact_paths)))
        if timeout_in_minutes:
            d["timeout_in_minutes"] = timeout_in_minutes
        if plugins:
            d["plugins"] = {}
            for plugin in list(plugins):
                d["plugins"][plugin.name] = plugin.args
        if self.priority is not None:
            d["priority"] = self.priority
        if self.manual_retry or self.automatic_retries:
            retry: Dict[str, Any] = {}
            if self.manual_retry:
                retry["manual"] = self.manual_retry.asdict()
            if self.automatic_retries:
                retry["automatic"] = [r.asdict() for r in self.automatic_retries]
            d["retry"] = retry

        if self.soft_fail:
            if isinstance(self.soft_fail, list):
                d["soft_fail"] = [{"exit_status": i} for i in self.soft_fail]
            elif self.soft_fail is True:
                d["soft_fail"] = True

        if self.concurrency:
            d["concurrency"] = self.concurrency
            if self.concurrency_group:
                d["concurrency_group"] = self.concurrency_group
        if self.parallelism:
            d["parallelism"] = self.parallelism
        if self.skip_reason:
            d["skip"] = self.skip_reason

        return d

    def __hash__(self) -> int:
        return hash(self.key)

    def __str__(self) -> str:
        return f"Command(key={self.key}, label={self.label})"


from typing import Union, Iterable

from .command import Command
from .group import Group
from .step import Step
from .target import Target
from .wait import Wait

from ..filter.filter import Filter


class Pipeline:
    """
    Methods for manipulating entire pipelines of steps, including
    the important role of rendering a de-duplicated properly
    ordered collection via steps(), or filtering based on
    targets and tags.
    """

    def __init__(self, steps: Iterable[Step]):
        self.items: list[Step] = list(steps)

    def filtered(self, filter: Filter) -> "Pipeline":
        """
        Filter the pipeline with the optional provided values and return a flattened
        list of steps with duplicate steps (via key) removed.
        """
        filtered: list[Step] = []
        for item in self.items:
            if isinstance(item, Group):
                filtered.append(item.filtered(filter))
            if isinstance(item, Step):
                if filter(item):
                    filtered.append(item)
        return Pipeline(filtered)

    @property
    def steps(self) -> list[Step]:
        # (1) Dependent Inclusion
        # Iterate through all steps and ensure all dependents are added to the list of steps,
        # even if they weren't in the list of steps added directly to the pipeline
        steps = self.items
        all_steps = steps[:]
        for step in steps:
            if isinstance(step, Group):
                all_steps += step.steps
        seen = set(all_steps)
        for step in all_steps:
            for dep in step.dependents:
                if dep not in seen:
                    seen.add(dep)
                    steps.append(dep)
                    all_steps.append(dep)

        # (2) Clean depends_on
        # Remove any depends_on keys for steps that have been removed via filtering
        keys = set([step.key for step in all_steps if isinstance(step, Command)])
        for step in all_steps:
            step.depends_on = [key for key in step.depends_on if key in keys]

        # (3) Remove empty groups
        cleaned: list[Step] = []
        for step in steps:
            if not isinstance(step, Group) or step.steps:
                cleaned.append(step)
        steps = cleaned

        # (4) Remove unnecessary Waits
        #     Remove runs of identical wait steps, and strip waits from the beginning/end
        #     of the pipeline
        last_step = None
        cleaned = []
        for step in steps:
            is_valid = True
            if isinstance(step, Wait):
                is_valid = last_step is not None and last_step != step
            last_step = step
            if is_valid:
                cleaned.append(step)
        if isinstance(cleaned[-1], Wait):
            cleaned.pop()
        steps = cleaned

        return steps

    def asdict(self) -> dict:
        d: dict = {"steps": [s.asdict() for s in self.steps]}
        return d

    def asyaml(self) -> str:
        import yaml

        return yaml.dump(self.asdict())



from typing import Iterable, List, Tuple, Union
from kitefly.filter.filter import Filter

from kitefly.util import generate_key

from .step import Step

class Group(Step):
  """
  Entity used to group multiple steps together, useful mostly for adding dependency relationships en-masse
  """
  def __init__(self, steps: Iterable[Step], *, key: str = "", label: str = "", **kwargs):
    super().__init__(**kwargs)
    self._steps: list[Step] = list(steps)
    self.label = label
    self.key = generate_key(label or 'Group')

  def __iadd__(self, value: Step) -> 'Group':
    if isinstance(value, Group):
      self._steps += value.steps
    else:
      self._steps.append(value)
    return self

  def __add__(self, value: Step) -> 'Group':
    group = Group(self._steps)
    group += value
    return group

  def __rshift__(self, dep: Union[Step, Iterable[Step]]) -> 'Group':
    for item in self._steps:
      item >> dep
    return self

  def __lshift__(self, dep: Union[Step, Iterable[Step]]) -> 'Group':
    for item in self._steps:
      item << dep
    return self

  def filtered(self, filter: Filter) -> 'Group':
    fs = [s for s in self._steps if filter(s)]
    return Group(fs, label=self.label)

  @property
  def steps(self) -> list[Step]:
    """
    Return a flattened list of steps
    """
    steps: list[Step] = []
    for step in self._steps:
      if isinstance(step, Group):
        steps += step.steps
      else:
        steps.append(step)
    return steps

  def asdict(self) -> dict:
    return {
      "group": self.label or "Group",
      "key": self.key,
      "steps": [s.asdict() for s in self.steps]
    }



from .step import Step

from .command import Command
from .group import Group
from .input import Option, Input, Block, TextField, SelectField
from .pipeline import Pipeline
from .plugin import Plugin
from .target import Target
from .trigger import BuildAttributes, Trigger
from .retry import AutomaticRetry
from .wait import Wait



class Plugin:
  """
  A Plugin configuration which may be applied to steps.

  See: https://buildkite.com/docs/plugins
  """
  def __init__(self, name: str, args: dict):
    self.name = name
    self.args = args


from typing import Any, Dict


class AutomaticRetry:
    """
    Configuration entry for an automatic retry on a Command Step, of which there may be
    one or more limit settings bound to particular exit codes. The '*' exit code (default)
    will match all exit codes.
    """

    def __init__(self, limit: int, exit_code: str = "*"):
        self.limit = limit
        try:
            exit_code_int = int(exit_code)
        except ValueError:
            exit_code_int = -1
        self.exit_code = exit_code_int if exit_code_int >= 0 else exit_code

    def asdict(self) -> dict:
        return {"exit_code": self.exit_code, "limit": self.limit}


class ManualRetry:
    """
    Configuration entry for manual retry on a Command Step.
    """

    def __init__(
        self, allowed: bool = True, permit_on_passed: bool = False, reason: str = ""
    ):
        self.allowed = allowed
        self.permit_on_passed = permit_on_passed
        self.reason = reason

    def asdict(self) -> dict:
        d: Dict[str, Any] = {
            "allowed": self.allowed,
        }
        if self.permit_on_passed:
            d["permit_on_passed"] = True
        if self.reason:
            d["reason"] = self.reason
        return d



import os
import re
from typing import Any, Iterable, Pattern, TypeVar, Union, cast

RE_NONID = re.compile(r"[^a-zA-Z0-9_]")
RE_MULTI_US = re.compile(r"__+")
T = TypeVar("T")
ST = TypeVar("ST")

KEY_COUNT = {}


def generate_key(name: str) -> str:
    """
    Given a display name, generate a global unique value suitable
    for use as a step key in Buildkite.
    """
    if not name:
        return ""
    norm = name.lower()
    for regex in (RE_NONID, RE_MULTI_US):
        norm = regex.sub("_", norm)
    if norm not in KEY_COUNT:
        KEY_COUNT[norm] = 1
        return norm
    suffix = f"__kf__{KEY_COUNT[norm]}"
    KEY_COUNT[norm] += 1
    return f"{norm}{suffix}"


def is_iterable(v: Any) -> bool:
    """
    Return True if the provided object is iterable.
    """
    try:
        _ = iter(v)
    except TypeError:
        return False
    return True


def as_iterable(v: Union[T, Iterable[T]]) -> Iterable[T]:
    """
    Convert passed object to a tuple.
    """
    if is_iterable(v):
        return cast(Iterable, v)
    else:
        return (cast(T, v),)


def glob(pattern: str) -> Pattern:
    """
    Return a compiled regular expression that will match full filepaths
    using the provided glob pattern.
    """
    escaped = re.escape(pattern)
    globkey = "__:GLOB:__"
    pattern = (
        escaped.replace("\\*\\*/", globkey)
        .replace("\\*\\*", globkey)
        .replace("\\*", f"[^{os.sep}]*")
        .replace(globkey, ".*")
    )
    return re.compile(pattern)



from typing import Any, Dict, Mapping, Optional

from .step import Step

class BuildAttributes:
  """
  Attributes used when triggering a build.
  """
  def __init__(
    self,
    message: str = "",
    commit: str = "",
    branch: str = "",
    meta_data: Optional[Mapping[str, str]] = None,
    env: Optional[Mapping[str, str]] = None,
  ):
    self.message = message
    self.commit = commit
    self.branch = branch
    self.meta_data = meta_data or {}
    self.env = env or {}

  def asdict(self) -> dict:
    d: Dict[str, Any] = {}
    if self.message:
      d["message"] = self.message
    if self.commit:
      d["commit"] = self.commit
    if self.branch:
      d["branch"] = self.branch
    if self.meta_data:
      d["meta_data"] = self.meta_data
    if self.env:
      d["env"] = self.env
    return d


class Trigger(Step):
  """
  A Step which triggers a build on another pipeline.

  See: https://buildkite.com/docs/pipelines/trigger-step
  """
  def __init__(
    self,
    pipeline: str,
    build: Optional[BuildAttributes] = None,
    label: str = "",
    asynchronous: bool = False,
    **kwargs
  ):
    super().__init__(**kwargs)
    self.pipeline = pipeline
    self.build = build
    self.label = label
    self.asynchronous = asynchronous

  def asdict(self) -> dict:
    d = super().asdict()
    d["trigger"] = self.pipeline
    if self.build:
      d["build"] = self.build.asdict()
    if self.label:
      d["label"] = self.label
    if self.asynchronous:
      d["async"] = True
    return d



from typing import Generator, List, Set, Iterable, Union, Pattern
from ..util import glob


class TargetPattern:
    GLOB_ALL = object()

    def __init__(self, pattern: Union[str, Pattern]):
        if isinstance(pattern, str):
            self.pattern = glob(pattern)
        else:
            self.pattern = pattern

    def matches(self, filepath: str) -> bool:
        return self.pattern.match(filepath) is not None

    def __len__(self) -> int:
        return len(self.pattern.pattern)

    def __lt__(self, comp: "TargetPattern") -> bool:
        return len(self) < len(comp)

    def __str__(self) -> str:
        return f"r/{self.pattern.pattern}/"


class Target:
    """
    A collection of file-matching patterns used to group sources for a single Build
    Target, which can be associated with one or more Step entities in the Pipeline.
    When base-branch comparison is used, each changed file in the list will be matched
    with the Target with the highest priority, or pattern length if the priorities are equal.
    This collection of Targets will be used to include/exclude steps in the pipeline.

    A Target can also depend on another Target, meaning that if A depends on B, and
    B is included in the list of matched Targets, then A will also be included.
    """

    def __init__(
        self,
        sources: Union[Iterable[Union[str, Pattern]], str, Pattern],
        priority: int = 0,
        name: str = "",
    ):
        if type(sources) == str or isinstance(sources, Pattern):
            self.patterns = [TargetPattern(sources)]
        elif type(sources) == list or type(sources) == tuple:
            self.patterns = [TargetPattern(p) for p in sources]
        self.depends_on: List[Target] = []
        self.name = name
        default_priority = getattr(self, "priority", 0)
        self.priority = priority or default_priority

    @property
    def dependencies(self) -> Set["Target"]:
        """
        Return distinct set of Targets in the dependency graph
        """
        deps: Set["Target"] = set()
        for target in self.depends_on:
            deps.add(target)
            deps |= target.dependencies
        return deps

    def _iterate_patterns(self) -> Generator[TargetPattern, None, None]:
        """
        Iterate recursively through all patterns in the depencency tree.
        """
        for pattern in self.patterns:
            yield pattern
        for target in self.depends_on:
            for pattern in target._iterate_patterns():
                yield pattern

    def matches(self, filepath: str) -> bool:
        """
        Returns True if the Target patterns or any of its dependency patterns match the provided filepath.
        """
        for pattern in self._iterate_patterns():
            if pattern.matches(filepath):
                return True
        return False

    @classmethod
    def src(cls, *sources: Union[str, Pattern]) -> "Target":
        return Target(sources=sources)

    def prio(self, priority: int) -> "Target":
        self.priority = priority
        return self

    def __rshift__(self, dep: "Target") -> "Target":
        self.depends_on.append(dep)
        return dep

    def __lt__(self, comp: "Target") -> bool:
        if self.priority != comp.priority:
            return self.priority < comp.priority
        return self.patterns < comp.patterns

    def __str__(self) -> str:
        plist = ",".join([str(p) for p in self.patterns])
        return f"Target(priority={self.priority},sources={plist})"



from typing import Any, Dict, List, Iterable, Optional, Union

from .target import Target
from ..util import as_iterable, is_iterable


class Step:
    """
    Generic root for pipeline steps containing common attributes such
    as 'if', 'branches', 'depends_on', and 'allow_dependency_failure'
    """

    _instance_count = 0

    def __init__(
        self,
        when: str = "",
        branches: str = "",
        allow_dependency_failure: bool = False,
        priority: Optional[int] = 0,
        tags: List[str] = None,
        targets: List[Target] = None,
        **kwargs
    ):
        self.__class__._instance_count += 1
        self.instance_serial = self.__class__._instance_count
        self.key = ""
        self.when = when
        self.depends_on: List[str] = []
        self.dependents: List[Step] = []
        self.branches = branches
        self.allow_dependency_failure = allow_dependency_failure
        self._tags = tags or []
        self._targets = targets or []
        self.properties: dict = kwargs
        self.priority = priority

    def classes(self) -> List[type]:
        """
        Return parent classes in reverse MRO, which is used to aggregate set or hash
        properties based on lineage.

        Example:
        cmd: Cokmand
        cmd.classes() -> [Step, Command]
        """
        classes = [cls for cls in self.__class__.__mro__ if cls is not object]
        classes.reverse()
        return classes

    def combined_parent_list(self, property: str) -> list:
        values: list = []
        print("CPL", property, self.classes())
        for cls in self.classes():
            cls_attr = cls.__dict__.get(property, [])
            if is_iterable(cls_attr):
                values += cls_attr
        return values

    def get_targets(self) -> List[Target]:
        """
        Return distinct list of Targets associated with this step.
        """
        return list(
            sorted(list(set(self._targets + self.combined_parent_list("targets"))))
        )

    def get_tags(self) -> list[str]:
        """
        Return distinct list of tags associated with this step.
        """
        return list(sorted(list(set(self._tags + self.combined_parent_list("tags")))))

    def asdict(self) -> dict:
        """
        Every Step class has an asdict() method to generate a plain dictionary mapping that can
        be used to generate serialized text formats, primarily YAML for Buildkite
        """
        d: Dict[str, Any] = {}
        if self.when:
            d["if"] = self.when
        if self.depends_on:
            d["depends_on"] = self.depends_on
        if self.branches:
            d["branches"] = self.branches
        if self.allow_dependency_failure:
            d["allow_dependency_failure"] = True
        if self.priority is not None:
            d["priority"] = self.priority
        if self.properties:
            d.update(self.properties)
        return d

    def __lshift__(self, deps: Union["Step", Iterable["Step"]]) -> "Step":
        """
        Declare a dependency relationship from the command step (A) with another step (B)
        that will become its dependent. The key of step A (keyA) will be added to B's
        depends_on list
        """
        for dep in as_iterable(deps):
            if not self.key:
                raise ValueError(
                    "Cannot add reverse dependency: self.key is not defined"
                )
            dep.depends_on.append(self.key)
            self.dependents.append(dep)
        return self

    def __rshift__(self, dep_on: Union["Step", Iterable["Step"]]) -> "Step":
        """
        Declare a dependency relationship from this step (A) on another step (B)
        Step B's key will be added to this step's depends_on list.
        """
        for parent in as_iterable(dep_on):
            if not parent.key:
                raise ValueError("Cannot depend on step: key is not defined")
            self.depends_on.append(parent.key)
            parent.dependents.append(self)
        return self

    def __eq__(self, step: Any) -> bool:
        if isinstance(step, Step):
            return self.asdict() == step.asdict()
        return False

    def __hash__(self) -> int:
        return self.instance_serial



from .step import Step

class Wait(Step):
  """
  Entity representing a "wait" step in a Buildkite pipeline

  See: https://buildkite.com/docs/pipelines/wait-step
  """
  def __init__(
    self,
    continue_on_failure: bool = False,
    **kwargs
  ):
    super().__init__(**kwargs)
    self.continue_on_failure = continue_on_failure

  def asdict(self) -> dict:
    d = super().asdict()
    d["wait"] = "~"
    if self.continue_on_failure:
      d["continue_on_failure"] = True
    return d


from typing import Any, Literal, Optional, Union

from .step import Step

class Option:
  def __init__(self, label: str, value: str):
    self.label = label
    self.value = value

class Field:
  def __init__(self, key: str):
    self.key = key

  def asdict(self) -> dict[str, Any]:
    return {"key": self.key}


class SelectField(Field):
  def __init__(
    self,
    key: str,
    name: str,
    options: list[Option],
    hint: str = "",
    required: bool = True,
    multiple: bool = False,
    default: Union[str, list[str]] = ""
  ):
    super().__init__(key)
    self.name = name
    self.options = options
    self.hint = hint
    self.required = required
    self.multiple = multiple
    self.default = default

  def asdict(self) -> dict:
    d = super().asdict()
    d.update({
      "select": self.name,
      "options": []
    })
    for option in self.options:
      d["options"].append({
        "label": option.label,
        "value": option.value,
      })
    if self.hint:
      d["hint"] = self.hint
    if not self.required:
      d["required"] = self.required
    if self.multiple:
      d["multiple"] = self.multiple
    if self.default:
      d["default"] = self.default
    return d

class TextField(Field):
  """
  A text field for Input step.
  """
  def __init__(
    self,
    key: str,
    name: str,
    required: bool = True,
    hint: str = "",
    default: str = ""
  ):
    super().__init__(key)
    self.name = name
    self.required = required
    self.hint = hint
    self.default = default

  def asdict(self) -> dict[str, Any]:
    d = super().asdict()
    d.update({
      "required": self.required,
      "text": self.name
    })
    if self.hint:
      d["hint"] = self.hint
    if self.default:
      d["default"] = self.default
    return d

BlockedState = Union[Literal["passed"], Literal["failed"], Literal["running"]]

class Input(Step):
  """
  A step that prompts the user for information that can be used in
  dependent builds

  See: https://buildkite.com/docs/pipelines/input-step
  """
  def __init__(
    self,
    label: str,
    prompt: str = "",
    fields: Optional[list[Field]] = None,
    blocked_state: BlockedState = "passed",
    **kwargs
  ):
    self.prompt = prompt
    self.label = label
    self.blocked_state = blocked_state
    self.fields = fields or []
    super().__init__(**kwargs)

  def asdict(self) -> dict:
    d = super().asdict()
    d["input"] = self.label
    if self.prompt:
      d["prompt"] = self.prompt
    if self.fields:
      d["fields"] = [f.asdict() for f in self.fields]
    if self.blocked_state:
      d["blocked_state"] = self.blocked_state
    return d


class Block(Input):
  """
  An Input Step that implicitly blocks any steps following it, functioning as a Wait and Input Step combined.

  See: https://buildkite.com/docs/pipelines/block-step
  """
  def asdict(self) -> dict:
    d = super().asdict()
    d["block"] = d["input"]
    del d["input"]
    return d


import functools
import os
import re
from subprocess import check_output, check_call
from typing import Dict, List

from .filter import Filter
from ..model.step import Step
from ..model.target import Target


class GitFilter(Filter):
    def __init__(self, base_branch: str = "") -> None:
        self.base_branch = base_branch or os.environ.get(
            "BUILDKITE_PULL_REQUEST_BASE_BRANCH", ""
        )
        self.match_cache: Dict[Target, bool] = {}
        super().__init__()

    def __call__(self, step: Step) -> bool:
        if not self.base_branch:
            return True
        for filepath in self._files_changed_since_branch(self.base_branch):
            for target in step.get_targets():
                if target.matches(filepath):
                    return True
        return False

    @functools.lru_cache()
    def _files_changed_since_branch(self, branch: str) -> List[str]:
        check_call(["git", "fetch", "origin", branch])
        git_root = check_output(
            ["git", "rev-parse", "--show-toplevel"], universal_newlines=True
        ).split(os.linesep)[0]
        lines = check_output(
            ["git", "diff", "--name-only", branch],
            universal_newlines=True,
            cwd=git_root,
        ).split(os.linesep)
        re_word_char = re.compile(r"\w")
        return [n.strip() for n in lines if re_word_char.match(n)]
